import numpy as np
from statsmodels.tsa.tsatools import lagmat

from .criticalValues import mapping as criticalValues_mapping


class Johansen(object):
    """Implementation of the Johansen test for cointegration.

    References:
        - Hamilton, J. D. (1994) 'Time Series Analysis', Princeton Univ. Press.
        - MacKinnon, Haug, Michelis (1996) 'Numerical distribution functions of
        likelihood ratio tests for cointegration', Queen's University Institute
        for Economic Research Discussion paper.

    """

    def __init__(self, x, model, k=1, trace=True,  significance_level=1):
        """
        :param x: (nobs, m) array of time series. nobs is the number of
        observations, or time stamps, and m is the number of series.

        :param k: The number of lags to use when regressing on the first
        difference of x.

        :param trace: Whether to use the trace or max eigenvalue statistic for
        the hypothesis testing. If False the latter is used.

        :param model: Which of the five cases in Osterwald-Lenum 1992 (or
        MacKinnon 1996) to use.
            - If set to 0, case 0 will be used. This case should be used if
            the input time series have no deterministic terms and all the
            cointegrating relations are expected to have 0 mean.
            - If set to 1, case 1* will be used. This case should be used if
            the input time series has neither a quadratic nor linear trend,
            but may have a constant term, and additionally if the cointegrating
            relations may have nonzero means.
            - If set to 2, case 1 will be used. This case should be used if
            the input time series have linear trends but the cointegrating
            relations are not expected to have linear trends.
            - If set to 3, case 2* will be used. This case should be used if
            the input time series do not have quadratic trends, but they and
            the cointegrating relations may have linear trends.
            - If set to 4, case 2 will be used. This case should be used if
            the input time series have quadratic trends, but the cointegrating
            relations are expected to only have linear trends.

        :param significance_level: Which significance level to use. If set to
        0, 90% significance will be used. If set to 1, 95% will be used. If set
        to 2, 99% will be used.

        """

        self.x = x
        self.k = k
        self.trace = trace
        self.model = model
        self.significance_level = significance_level

        if trace:
            key = "TRACE_{}".format(model)
        else:
            key = "MAX_EVAL_{}".format(model)

        critical_values_str = criticalValues_mapping[key]

        select_critical_values = np.array(
            critical_values_str.split(),
            float).reshape(-1, 3)

        self.critical_values = select_critical_values[:, significance_level]

    def mle(self):
        """Obtain the cointegrating vectors and corresponding eigenvalues.

        Maximum likelihood estimation and reduced rank regression are used to
        obtain the cointegrating vectors and corresponding eigenvalues, as
        outlined in Hamilton 1994.

        :return: The possible cointegrating vectors, i.e. the eigenvectors
        resulting from maximum likelihood estimation and reduced rank
        regression, and the corresponding eigenvalues.

        """

        # Regressions on diffs and levels of x. Get regression residuals.

        # First differences of x.
        x_diff = np.diff(self.x, axis=0)

        # Lags of x_diff.
        x_diff_lags = lagmat(x_diff, self.k, trim='both')

        # First lag of x.
        x_lag = lagmat(self.x, 1, trim='both')

        # Trim x_diff and x_lag so they line up with x_diff_lags.
        x_diff = x_diff[self.k:]
        x_lag = x_lag[self.k:]

        # Include intercept in the regressions if self.model != 0.
        if self.model != 0:
            ones = np.ones((x_diff_lags.shape[0], 1))
            x_diff_lags = np.append(x_diff_lags, ones, axis=1)

        # Include time trend in the regression if self.model = 3 or 4.
        if self.model in (3, 4):
            times = np.asarray(range(x_diff_lags.shape[0])).reshape((-1, 1))
            x_diff_lags = np.append(x_diff_lags, times, axis=1)

        # Residuals of the regressions of x_diff and x_lag on x_diff_lags.
        try:
            inverse = np.linalg.pinv(x_diff_lags)
        except:
            print("Unable to take inverse of x_diff_lags.")
            return None

        u = x_diff - np.dot(x_diff_lags, np.dot(inverse, x_diff))
        v = x_lag - np.dot(x_diff_lags, np.dot(inverse, x_lag))

        # Covariance matrices of the residuals.
        t = x_diff_lags.shape[0]
        Svv = np.dot(v.T, v) / t
        Suu = np.dot(u.T, u) / t
        Suv = np.dot(u.T, v) / t
        Svu = Suv.T

        try:
            Svv_inv = np.linalg.inv(Svv)
        except:
            print("Unable to take inverse of Svv.")
            return None
        try:
            Suu_inv = np.linalg.inv(Suu)
        except:
            print("Unable to take inverse of Suu.")
            return None

        # Eigenvalues and eigenvectors of the product of covariances.
        cov_prod = np.dot(Svv_inv, np.dot(Svu, np.dot(Suu_inv, Suv)))
        eigenvalues, eigenvectors = np.linalg.eig(cov_prod)

        # Normalize the eigenvectors using Cholesky decomposition.
        evec_Svv_evec = np.dot(eigenvectors.T, np.dot(Svv, eigenvectors))
        cholesky_factor = np.linalg.cholesky(evec_Svv_evec)
        try:
            eigenvectors = np.dot(eigenvectors,
                                  np.linalg.inv(cholesky_factor.T))
        except:
            print("Unable to take the inverse of the Cholesky factor.")
            return None

        # Ordering the eigenvalues and eigenvectors from largest to smallest.
        indices_ordered = np.argsort(eigenvalues)
        indices_ordered = np.flipud(indices_ordered)
        eigenvalues = eigenvalues[indices_ordered]
        eigenvectors = eigenvectors[:, indices_ordered]

        return eigenvectors, eigenvalues

    def h_test(self, eigenvalues, r):
        """Carry out hypothesis test.

        The null hypothesis is that there are at most r cointegrating vectors.
        The alternative hypothesis is that there are at most m cointegrating
        vectors, where m is the number of input time series.

        :param eigenvalues: The list of eigenvalues returned from the mle
        function.

        :param r: The number of cointegrating vectors to use in the null
        hypothesis.

        :return: True if the null hypothesis is rejected, False otherwise.

        """

        nobs, m = self.x.shape
        t = nobs - self.k - 1

        if self.trace:
            m = len(eigenvalues)
            statistic = -t * np.sum(np.log(np.ones(m) - eigenvalues)[r:])
        else:
            statistic = -t * np.sum(np.log(1 - eigenvalues[r]))

        critical_value = self.critical_values[m - r - 1]

        if statistic > critical_value:
            return True
        else:
            return False

    def johansen(self):
        """Obtain the possible cointegrating relations and numbers of them.

        See the documentation for methods mle and h_test.

        :return: The possible cointegrating relations, i.e. the eigenvectors
        obtained from maximum likelihood estimation, and the numbers of
        cointegrating relations for which the null hypothesis is rejected.

        """

        nobs, m = self.x.shape

        try:
            eigenvectors, eigenvalues = self.mle()
        except:
            print("Unable to obtain possible cointegrating relations.")
            return None

        rejected_r_values = []
        for r in range(m):
            if self.h_test(eigenvalues, r):
                rejected_r_values.append(r)

        return eigenvectors, rejected_r_values



# -*- coding: utf-8 -*-
"""
Created on Thu Jul  4 11:02:08 2019

@author: Administrator
"""

#@title
#!pip install quandl
#import quandl
# Request: allow HTTP-post API calls
import requests
# Json: to format as json calls to API
import json 
import numpy as np
import matplotlib.pyplot as plt
from statsmodels.tsa.stattools import adfuller
import itertools
# Json: to format as json calls to API

#@title
#!pip install johansen
# Request: allow HTTP-post API calls
import requests
# Json: to format as json calls to API
import json 
# Pandas: to plot the timeseries
import pandas as pd
# Datetime: to convert different date formats
import datetime as dt
# Import the Time Series library
import statsmodels.tsa.stattools as ts


import numpy as np
from numpy import cumsum, log, polyfit, sqrt, std, subtract
#@title
#from johansen import Johansen
from .johan import Johansen

import matplotlib.pyplot as plt
import operator as op



# API_ENDPOINT = "https://markovsimulator.azurewebsites.net/api/MarkovCalibrator_01?code=UFYL9CcFRERnVbZJ3ELdRRY7u59vD3BgONzaGLNEOekTfKf0Ug9Bfw=="
API_ENDPOINT = "https://gmaatest.azurewebsites.net/api/HttpTriggerCSharp1?code=COXmk9ou6Mr8/dUCYj045SaeNm6jCkPzJ3ilU/dCiTvNcFNQax4BMQ=="


#@title
# Call Quandl to get all the data
def readQuandl(FXid, QUANDL_API_KEY):
  QUAND_FX_prefix = "https://www.quandl.com/api/v3/datasets/"
  QUAND_FX_suffix = ".json?api_key="+QUANDL_API_KEY

  #print (QUAND_FX_prefix+FXid+QUAND_FX_suffix)
  q = requests.get(url = QUAND_FX_prefix+FXid+QUAND_FX_suffix)
  
  q_json = q.json() 
  q_data = q_json['dataset']['data']

  # Compute derived series from Quandl Data
  dates = []
  values = []
  q_len = len(q_data)
  column_len = len(q_data[0])
  column = column_len-1
  columns = q_json['dataset']['column_names']
  #print (columns)


  # The following could be easily done using Quandl's
  # python API - but I keep it the 'hard way' to allow
  # compatibility with Google Colaboratory
  df = pd.DataFrame.from_records(q_data, columns = columns)
  df['Date'] = pd.to_datetime(df['Date'])
  df.set_index('Date', inplace=True)
  return df


def ouCalibrateMarkov(df_ts, column_name = None, calType = "MeanReverting"):
    # if no column name is given takes only the first one.
    # to send data to the Azure API we need to:
    # - sort the dates increasing
    # - remove NaNs
    # - put dates in string format readable by Azure
    
    # Time Series seed
    json_ts =  {
       "calType" : "",
        "seriesName" : 'seriesName',
        "colName" : "Value",
      "dates": [],  
      "vals" : []
    }

    panda_ts = df_ts.copy()    
    panda_ts.sort_index(inplace = True)
    panda_ts['dates_'] = panda_ts.index
    panda_ts = panda_ts.dropna()

    json_ts['dates'] = panda_ts['dates_'].apply(lambda x: x.strftime('%Y-%m-%d')).tolist()
    if not column_name:
        json_ts['vals'] = panda_ts[panda_ts.columns[0]].tolist()
    else:
        json_ts['vals'] = panda_ts[column_name].tolist()
    json_ts['calType'] = calType
    r = requests.post(url = API_ENDPOINT, json = json_ts)
    r_json = r.json()
    theta = r_json['calibrationResult']['Lambda']
    mu = r_json['calibrationResult']['Mu']
    half_life = -np.log(0.5) * 365.25 / theta
    r_json['calibrationResult']['HalfLife'] = half_life
    sigma = r_json['calibrationResult']['Sigma']
    #r_json['calibrationResult']['Std_band'] = np.sqrt(sigma * sigma / (2 * theta))
    std_band = np.sqrt(sigma * sigma / (2 * theta))
    
    #return r_json
    return (mu,sigma,half_life, std_band)  


def ouCalibrateML(df,delta):
  S = df.values
  from numpy import sum,power,sqrt,log,exp
  n = len(S)-1

  Sx = sum(S[0:-1])
  Sy = sum(S[1: ])
  Sxx = sum(power(S[0:-1],2))
  Sxy = sum( S[0:-1] * S[1: ])
  Syy = sum(power(S[1: ],2))

  mu  = (Sy*Sxx - Sx*Sxy) / ( n*(Sxx - Sxy) - (power(Sx,2) - Sx*Sy) )

  lambda_ = -log( (Sxy - mu*Sx - mu*Sy + n*power(mu,2)) / (Sxx -2*mu*Sx + n*power(mu,2)) ) / delta
  a = exp(-lambda_*delta)
  sigmah2 = (Syy - 2*a*Sxy + power(a,2)*Sxx - 2*mu*(1-a)*(Sy - a*Sx) + n*power(mu,2)*power(1-a,2))/n
  sigma = sqrt(sigmah2*2*lambda_/(1-power(a,2)))

  #half_life = -np.log(0.5) * 365.25 / lambda_
  half_life = -np.log(0.5) / lambda_
  
  std_band = np.sqrt(sigma * sigma / (2 * lambda_))

  return (mu,sigma,half_life, std_band)   



#
#TimeSeries = cryptos[assets].dropna()
#TimeSeriesTraining = TimeSeries.loc['2014-01-01' : '2018-03-01']
#https://github.com/iisayoo/johansen
#http://qed.econ.queensu.ca/working_papers/papers/qed_wp_1227.pdf
def findMeanRevertingSignals(timeSeries, significance_level = 1):
    #x = timeSeries.as_matrix()
    x = timeSeries.as_matrix()
    x_centered = x - np.mean(x, axis=0)

    # model:
    # If set to 2, case 1 will be used. This case should be used if
    #        the input time series have linear trends but the cointegrating
    #        relations are not expected to have linear trends.
    #   significance_level: 0, 90% significance will be used. 
    #                       1, 95% will be used. If set
    #                       2, 99% will be used.
    jhnsn = Johansen(x_centered, model=2, significance_level=significance_level)
    try:
      eigenvectors, r = jhnsn.johansen()
      #print (r)

      mean_rev = [False]
      for k in r:

          j_ev = eigenvectors[:, k] 


          bb = pd.DataFrame(timeSeries.dot(j_ev))
          
          #r_json = MarkovCalibration(bb)
          #half_life = r_json['calibrationResult']['HalfLife']
          #sigma = r_json['calibrationResult']['Sigma']

          #mu = r_json['calibrationResult']['Mu']
          #theta_ = r_json['calibrationResult']['Lambda']    
          #std_band = r_json['calibrationResult']['Std_band']
          ##cc = bb - mu
          ##dd = TimeSeries * j_ev
          #mean_rev.append([list(timeSeries.columns), j_ev, [mu, sigma, half_life, std_band]]) 
          mu, sigma, half_life, std_band = ouCalibrateML(bb, 1)
          mean_rev.append([list(timeSeries.columns), j_ev, [mu, sigma, half_life, std_band]]) 
          
          #mean_rev.append([list(timeSeries.columns), j_ev]) 
          mean_rev[0] = True
    except:
      #print ("Numerical error")
      return [False]

    return mean_rev

def plotMulti(data, cols=None, spacing=.1, **kwargs):

    from pandas import plotting

    # Get default color style from pandas - can be changed to any other color list
    if cols is None: cols = data.columns
    if len(cols) == 0: return
    colors = getattr(getattr(plotting, '_style'), '_get_standard_colors')(num_colors=len(cols))

    # First axis
    ax = data.loc[:, cols[0]].plot(label=cols[0], color=colors[0], **kwargs)
    ax.set_ylabel(ylabel=cols[0])
    lines, labels = ax.get_legend_handles_labels()

    for n in range(1, len(cols)):
        # Multiple y-axes
        ax_new = ax.twinx()
        ax_new.spines['right'].set_position(('axes', 1 + spacing * (n - 1)))
        data.loc[:, cols[n]].plot(ax=ax_new, label=cols[n], color=colors[n % len(colors)])
        ax_new.set_ylabel(ylabel=cols[n])

        # Proper legend position
        line, label = ax_new.get_legend_handles_labels()
        lines += line
        labels += label

    ax.legend(lines, labels, loc=0)
    return ax




def scatterLongShort(TimeSeries, TrainingLen, j_ev, ax):
  #j_short = j_ev * (j_ev <= 0 )
  #j_long = j_ev * (j_ev > 0 )
  if (all(j_ev >=0) or all(j_ev <=0)):
    j_filt = j_ev * 0
    j_filt[0] = 1
    j_filt2 = 1 - j_filt
    j_short = j_ev * (j_filt )
    j_long = j_ev * (j_filt2 )
    #print ('All one way')
  else:
    j_short = j_ev * (j_ev <= 0 )
    j_long = j_ev * (j_ev > 0 )
    
  
  #print (j_short, j_long)

  scat = -pd.DataFrame(TimeSeries.dot(j_short))
  scat['long'] = pd.DataFrame(TimeSeries.dot(j_long))
  #print (scat.head())
  #ax = plt.subplot()
  color = ['b'] * (len(TimeSeries) - TrainingLen) + ['r']*(TrainingLen)
  ax.scatter(scat[0], scat['long'], color=color)
  #scat.plot(kind = 'scatter', x = 0, y = 'long' )

  port_usd = TimeSeries * j_short + TimeSeries * j_long
  #plt.show()
  #print (port_usd[-1:])
  
  

def ncr(n, r):
    r = min(r, n-r)
    numer = reduce(op.mul, xrange(n, n-r, -1), 1)
    denom = reduce(op.mul, xrange(1, r+1), 1)
    return numer//denom

def plotMrBands(ts,LastCalibrationDay,mu,std_band, ax2, half_life = None, frequency = 'D'):
    ts.loc[:LastCalibrationDay].plot(ax = ax2)
    lastValue = ts.iloc[-1]
    #bb.loc[:'2018-1-1'].plot(ax = ax2)
    ts.loc[LastCalibrationDay:].plot(ax = ax2, color = 'r')
    #mu = item[1][2][0]
    #std_band = item[1][2][3]
    #cc['mu'] = mu / maxEurWeight
    #cc['2p_std'] = cc['mu'] + 2 * std_band / maxEurWeight
    #cc['2m_std'] = cc['mu'] - 2 * std_band / maxEurWeight
    ts['mu'] = mu
    ts['2p_std'] = ts['mu'] + 2 * std_band
    ts['2m_std'] = ts['mu'] - 2 * std_band

    ts[['mu','2p_std', '2m_std']].plot(ax = ax2, style = ['y-.', 'g-.', 'g-.'])
    if ((half_life is not None)):
        fwd = forwardPath(ts.index.max(),lastValue,mu,std_band,half_life,frequency )
        fwd[['fwd_mu','fwd_2p_std','fwd_2m_std']].plot(ax = ax2)
# Using the formulas to compute the forward expected mean and standard devition at some time
# t in the future (today being time '0', and expressed in terms of a year being one unit)
def OUmu_t(currentValue, mu, theta, t):
  return mu + (currentValue - mu)*np.exp(-theta * t)

def OUsigma_t(sigma, theta, t):
  return np.sqrt (  ( 1 - np.exp(-2 * theta * t)) * sigma * sigma / (2 * theta) )    

def forwardPath(last_date, last_value, mean, std_band, half_life, frequency):
    # Draw the forward path 6 x half life as mentioned above
    rng = pd.date_range(last_date, 
                        periods=int(half_life*6), 
                        freq=frequency)
    theta = -np.log(0.5)/half_life
    #std_band = np.sqrt(sigma * sigma / (2 * lambda_))
    sigma = np.sqrt(std_band * std_band * 2 * theta)
    fwd = pd.Series(rng).to_frame('fwd')
    fwd['delta'] = fwd.index
    fwd['fwd_mu'] = fwd['delta'].apply(lambda x: OUmu_t(last_value, mean, theta, x ))
    fwd['fwd_std'] = fwd['delta'].apply(lambda x: OUsigma_t(sigma, theta, x ))
    fwd['fwd_2p_std'] = fwd['fwd_mu'] + 2 * fwd['fwd_std']
    fwd['fwd_2m_std'] = fwd['fwd_mu'] - 2 * fwd['fwd_std']
    fwd = fwd.set_index('fwd')
    return(fwd)
    
# useful sites for pip installation:
# https://packaging.python.org/tutorials/packaging-projects/#uploading-your-project-to-pypi
# https://flummox-engineering.blogspot.com/2017/08/pypi-upload-failed-403-invalid-or-non-existent-authentication-information.html
    






from setuptools import setup, find_packages


#import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# This call to setup() does all the work
setup( 
    name="mean-reversion-util",
    version="0.0.1",
    description="Time Series mean reversion utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gjlr2000/johansen3.git",
    author="Gerardo Lemus",
    author_email="gerardo@alum.mit.edu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ],
    packages=find_packages(),
    install_requires=[
          'scipy>=0.18.0',
          'statsmodels>=0.6.1',
          'numpy>=1.11.1',
          'pandas>=0.18.1',
          #'binance>=0.7.1',            
      ],
    include_package_data=True,

)


# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 10:26:18 2019

@author: Administrator
"""

# importing the requests library 
import requests 
import csv
import urllib.request
import codecs
import pandas as pd
import datetime as dt
#!pip install mean-reversion-utilities
from .utilities import readQuandl
#!pip install python-binance
from binance.client import Client
import pandas as pd
from datetime import datetime




def portReadCryptoData(assets,prefix,suffix,column):
  firstTime = True 
  for asset in assets: 
    url = prefix+asset+suffix
    r = requests.get(url = url, verify = False) 

    text = r.iter_lines()
    reader = csv.reader(codecs.iterdecode(text, 'utf-8'))


    from io import StringIO

    data = StringIO(r.text)

    df = pd.read_csv(data,skiprows = [0], header=0)
    df.set_index('Date', inplace = True)
    
    if firstTime:
      
      timeSeries = pd.DataFrame(df[column])
      timeSeries.columns = [asset]
      firstTime = False
    else:
      timeSeries[asset] = df[column]
      #timeSeries = pd.concat([timeSeries, df[column]], axis = 1, sort=True)
      
  return timeSeries  

def portReadBinance(cryptos,column,timeInterval,start_date,end_date,api_key,api_secret):
  
  client = Client(api_key, api_secret)
  first_time = True
  for crypto in cryptos:

    klines = client.get_historical_klines(crypto, timeInterval, start_date, end_date)

    df = pd.DataFrame(klines)
    try:
      df.columns = ['openTime',crypto,'high','low','close','volume','closeTime','quoteAssetVolume','numberOfTrades','takerbuyBaseAssetVolume','takerBuyQuoteAssetVolume','ignore']
      df['date'] = pd.to_datetime(pd.Series(df['openTime']),unit='ms')
      df.set_index('date', inplace=True)
      df[crypto] = pd.to_numeric(df[crypto])
      #print ('loading '+crypto)
      if first_time:
        cryptos_df = pd.DataFrame(df[crypto])
        first_time = False
      else:
          cryptos_df[crypto] = df[crypto]
          #cryptos_df = pd.concat([cryptos_df, df[crypto]], axis = 1)

    except:
      dummy_line = 0
      #print ('problem with '+crypto)
  return cryptos_df    


def portReadQuandl(assets,Crypto_flag, QUANDL_API_KEY):

  firstAsset = True
  
  if Crypto_flag:
    source = 'BITFINEX'
    column = 'Mid'

  else:
    source = 'ECB'
    column = 'Value'


  for asset in assets:
    q_ticker = source+'/'+asset.upper()

    #print (q_ticker)
    if True:
    #  try:
      #df = quandl.get(q_ticker)
      df = readQuandl(q_ticker, QUANDL_API_KEY)
      df.rename(columns={column:asset.upper()}, inplace=True)
      if firstAsset:
        timeS = df[asset.upper()]
        firstAsset = False
      else:
        timeS = pd.concat([timeS, df[asset.upper()]], axis = 1)

  #  except:
  #    print (asset.upper()+' not in ' + source)

#df.head()
  return timeS 
